#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#define WIDTH 500
#define HEIGHT 600
#define BOMBFALLSPEED 5
#define BOMBDROPDELAY 2000
#define NUKEEXPANDSPEED 50
#define STARTRADIUS 0
#define MAXRADIUS 50
#define MAXCITIES 6
#define CITYHEIGHT 50
#define CITYWIDTH (WIDTH/((MAXCITIES+1)*2))
#define GAMETIME 60000 // 1 minute

struct game
	{
	int numCities;	// the number of cities that the player has
	int cities[MAXCITIES];
	GdkRectangle cityBoxes[MAXCITIES];
	int bombDropperID; // the regular bomb dropper
	};
typedef struct game game;

struct bomb
	{
	float x,y;	// current pos
	float xd,yd;	// current direction
	int tickerID;	// id of ticker function
	GtkWidget *darea;
	GSList *list;	// the list that I'm in
	GSList *nukes;	// the enemy list
	game *theGame;	// current game state
	};
typedef struct bomb bomb;

struct nuke
	{
	int x,y;	// position
	int radius;	// current radius
	int rd;		// direction radius is expanding in, 1 or -1
	int tickerID;	// id of ticker function
	GtkWidget *darea;
	GSList *list;	// the list that I'm in
	game *theGame;	// current game state
	};
typedef struct nuke nuke;

GtkWidget *createDrawWindow(int xsize,int ysize,int xoff,int yoff);
GtkWidget *createDrawArea(GtkWidget *window,GtkWidget *vbox,int xsize,int ysize);
GtkWidget *createVBox(GtkWidget *window);
gint expose_event (GtkWidget *widget, GdkEventExpose *event);
gint configure_event (GtkWidget *widget, GdkEventConfigure *event);
gint button_release_event(GtkWidget *widget, GdkEventButton *event);
void quit(void);
bomb *newBomb(void);
void destroyBomb(bomb *b);
void initBomb(bomb *b);
gint moveBomb(gpointer *data);
gint timeout_cb(gpointer *data);
void repaint(GtkWidget *w);
int checkEdgeCollision(bomb *b);
gint dropBomb(gpointer *data);
nuke *newNuke(void);
void destroyNuke(nuke *n);
int checkNukeBombCollision(GSList *nukes, bomb *b);
int insideCircle(int x1,int y1,int x2,int y2,int r);
void drawCities(GtkWidget *darea,game *theGame);
int checkCityCollision(bomb *b);
gint expandNuke(gpointer *data);
gint stopPlay(gpointer *data);

GdkPixmap *pixmap=NULL;

int main(int argc, char *argv[])
{
GtkWidget *window,*darea,*vbox;
GSList *bombs,*nukes;
int gameTimerID;
game theGame;
int i;

for (i=0;i<MAXCITIES;i++)
	{
	theGame.cities[i] = 1;
	theGame.cityBoxes[i].x = CITYWIDTH*(i+1)*2;
	theGame.cityBoxes[i].y = HEIGHT-CITYHEIGHT;
	theGame.cityBoxes[i].height = CITYHEIGHT;
	theGame.cityBoxes[i].width = CITYWIDTH;
	}
	
bombs = g_slist_alloc();
nukes = g_slist_alloc();

gtk_init(&argc,&argv);

window = createDrawWindow(WIDTH,HEIGHT,20,20);
vbox = createVBox(window);
darea = createDrawArea(window,vbox,WIDTH,HEIGHT);

gtk_object_set_data(GTK_OBJECT(window),"darea",darea);
gtk_object_set_data(GTK_OBJECT(window),"bombs",bombs);
gtk_object_set_data(GTK_OBJECT(window),"nukes",nukes);
gtk_object_set_data(GTK_OBJECT(window),"theGame",&theGame);

gtk_widget_show(window);

theGame.bombDropperID = gtk_timeout_add(BOMBDROPDELAY,(GtkFunction)dropBomb,(gpointer)window);
gameTimerID = gtk_timeout_add(GAMETIME,(GtkFunction)stopPlay,(gpointer)window);

drawCities(darea,&theGame);

gtk_main();

g_slist_free(bombs);
g_slist_free(nukes);

return 0;
}

gint stopPlay(gpointer *data)
{
game *theGame;
int num,i;
GdkGC *gc;
GtkWidget *darea;
GSList *bombs,*nukes;
bomb *b;
nuke *n;

theGame = gtk_object_get_data(GTK_OBJECT(data),"theGame");
darea = gtk_object_get_data(GTK_OBJECT(data),"darea");
bombs = (GSList *)gtk_object_get_data(GTK_OBJECT(data),"bombs");
nukes = (GSList *)gtk_object_get_data(GTK_OBJECT(data),"nukes");
gtk_timeout_remove(theGame->bombDropperID);

// remove all of the moving bombs from the air
for (i=1;i<g_slist_length(bombs);i++)
        {
        b = (bomb *)g_slist_nth_data(bombs,i);
        gtk_timeout_remove(b->tickerID);
        }

// stop all of the nukes
for (i=1;i<g_slist_length(nukes);i++)
        {
        n = (nuke *)g_slist_nth_data(nukes,i);
        gtk_timeout_remove(n->tickerID);
        }


num=0;
for (i=0;i<MAXCITIES;i++) if (theGame->cities[i]) num++;

fprintf(stderr,"\n%d cities left",num);
gc = gdk_gc_new(darea->window);
gdk_gc_set_foreground(gc,&(darea->style->white));
gdk_draw_rectangle(pixmap,gc,1,0,0,WIDTH,HEIGHT);
gdk_draw_rectangle(darea->window,gc,1,0,0,WIDTH,HEIGHT);
drawCities(darea,theGame);

return FALSE;
}

gint dropBomb(gpointer *data)
{
GSList *bombs;
bomb *b;
GtkWidget *darea;
GSList *nukes;
game *theGame;

darea = gtk_object_get_data(GTK_OBJECT(data),"darea");
bombs = (GSList *)gtk_object_get_data(GTK_OBJECT(data),"bombs");
nukes = (GSList *)gtk_object_get_data(GTK_OBJECT(data),"nukes");
theGame = gtk_object_get_data(GTK_OBJECT(data),"theGame");

b = newBomb();
initBomb(b);
b->theGame = theGame;
b->darea = darea;
b->nukes = nukes;
b->list = bombs;
b->tickerID = gtk_timeout_add(BOMBFALLSPEED,(GtkFunction)moveBomb,(gpointer)b);
g_slist_append(bombs,(gpointer)b);

return TRUE;
}

bomb *newBomb(void)
{
bomb *b;
if ( (b=(bomb *)malloc(sizeof(bomb))) == NULL ) exit(-1);
return b;
}

nuke *newNuke(void)
{
nuke *n;
if ( (n=(nuke *)malloc(sizeof(nuke))) == NULL ) exit(-1);
return n;
}

void destroyBomb(bomb *b)
{
free(b);
}

void destroyNuke(nuke *n)
{
free(n);
}

void initBomb(bomb *b)
{
b->x = rand()%WIDTH;
b->y = 0;

b->xd = (((rand()%11)-5)/10.0)*2.0;
b->yd = ((rand()%11)/10.0)*2.0;

if (b->xd == 0.0 || b->yd == 0.0) initBomb(b);
}

gint moveBomb(gpointer *data)
{
bomb *b;
GdkGC *gc;
GSList *bombs;
int hitCity=0;

b = (bomb *)data;

bombs = b->list;

b->x+=b->xd;
b->y+=b->yd;

gc = gdk_gc_new(b->darea->window);
gdk_draw_rectangle(pixmap,gc,0,b->x-1,b->y-1,2,2);
gdk_draw_rectangle(b->darea->window,gc,0,b->x-1,b->y-1,2,2);

if (checkCityCollision(b))
	{
	hitCity=1;
	}

if (checkEdgeCollision(b) || checkNukeBombCollision(b->nukes,b) || hitCity)
	{
	g_slist_remove(bombs,b);
	gtk_timeout_remove(b->tickerID);
	destroyBomb(b);
	}

return TRUE;
}

int checkCityCollision(bomb *b)
{
game *theGame;
int i;
nuke *n;

theGame = b->theGame;

for (i=0;i<MAXCITIES;i++)
	{
	if (theGame->cities[i]) // if still standing
		{
		if (b->x >= theGame->cityBoxes[i].x && b->x <= theGame->cityBoxes[i].x+theGame->cityBoxes[i].width && b->y >= theGame->cityBoxes[i].y && b->y <= theGame->cityBoxes[i].y+theGame->cityBoxes[i].height)
			{ // hit city i
			// blow it up a bit.
			theGame->cities[i] = 0;
			n = newNuke();
			n->theGame = theGame;
			n->darea = b->darea;
			n->x = b->x;
			n->y = b->y;
			n->radius = STARTRADIUS;
			n->rd = 1; // radius is expanding
			n->list = b->nukes;
			n->tickerID = gtk_timeout_add(NUKEEXPANDSPEED,(GtkFunction)expandNuke,(gpointer)n);
			g_slist_append(b->nukes,(gpointer)n);
			return 1;
			}
		}
	}
return 0;
}

void drawCities(GtkWidget *darea,game *theGame)
{
GdkGC *gc;
int i;

gc = gdk_gc_new(darea->window);

for (i=0;i<MAXCITIES;i++)
	{
	if (theGame->cities[i])
		{
		gdk_draw_rectangle(pixmap,gc,1,theGame->cityBoxes[i].x,theGame->cityBoxes[i].y,theGame->cityBoxes[i].width,theGame->cityBoxes[i].height);
		gdk_draw_rectangle(darea->window,gc,1,theGame->cityBoxes[i].x,theGame->cityBoxes[i].y,theGame->cityBoxes[i].width,theGame->cityBoxes[i].height);
		}
	}
}


int checkEdgeCollision(bomb *b)
{
if (b->x>=WIDTH || b->x<=0 || b->y<=0 || b->y>=HEIGHT) return 1;
else return 0;
}

int checkNukeBombCollision(GSList *nukes, bomb *b)
{
// has this bomb hit any nukes?
int i;
nuke *n;

for (i=1;i<g_slist_length(nukes);i++)
	{
	n = (nuke *)g_slist_nth_data(nukes,i);
	if (insideCircle(b->x,b->y,n->x,n->y,n->radius)) return 1;
	}
return 0;
}

int insideCircle(int x1,int y1,int x2,int y2,int r)
{
// is x1,y1 in x2,y2,r?
return (((x1-x2)*(x1-x2))+((y1-y2)*(y1-y2))<=(r*r));
}

gint expandNuke(gpointer *data)
{
nuke *n;
GdkGC *gc;
GSList *nukes;

n = (nuke *)data;

nukes = n->list;
n->radius+=n->rd;
if (n->radius>=MAXRADIUS) n->rd=-1;

gc = gdk_gc_new(n->darea->window);

if (n->rd==1)
gdk_gc_set_foreground(gc,&(n->darea->style->black));
else
gdk_gc_set_foreground(gc,&(n->darea->style->white));

gdk_draw_arc(pixmap,gc,0,n->x-n->radius,n->y-n->radius,n->radius*2,n->radius*2,0,360*64);
gdk_draw_arc(n->darea->window,gc,0,n->x-n->radius,n->y-n->radius,n->radius*2,n->radius*2,0,360*64);

if (n->radius <= 0 && n->rd == -1)
	{
	g_slist_remove(nukes,n);
 	gtk_timeout_remove(n->tickerID);
 	destroyNuke(n);
	}

return TRUE;
}

void quit(void)
{
exit(-1);
}

GtkWidget *createDrawWindow(int xsize,int ysize,int xoff,int yoff)
{
GtkWidget *window;  
window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
gtk_window_set_policy((GtkWindow *)window,0,0,1);
gtk_widget_set_uposition(window,xoff,yoff);
gtk_window_set_title((GtkWindow *)window,"Missile");
gtk_signal_connect(GTK_OBJECT(window),"destroy",GTK_SIGNAL_FUNC (quit), NULL);
return window;
} // end create window to draw in.

GtkWidget *createDrawArea(GtkWidget *window,GtkWidget *vbox,int xsize,int ysize)
{
GtkWidget *drawing_area;
drawing_area = gtk_drawing_area_new ();

gtk_object_set_data(GTK_OBJECT(drawing_area),"window",window);

gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), xsize, ysize);
gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);
// Signals used to handle backing pixmap
gtk_signal_connect(GTK_OBJECT (drawing_area), "expose_event",
                  (GtkSignalFunc) expose_event, NULL);
gtk_signal_connect(GTK_OBJECT(drawing_area),"configure_event",
                  (GtkSignalFunc) configure_event, NULL);
gtk_signal_connect(GTK_OBJECT(drawing_area),"button_release_event",
	          (GtkSignalFunc) button_release_event, NULL);
gtk_widget_set_events(drawing_area, GDK_EXPOSURE_MASK
                      | GDK_LEAVE_NOTIFY_MASK
                      | GDK_BUTTON_PRESS_MASK
                      | GDK_BUTTON_RELEASE_MASK
                      | GDK_POINTER_MOTION_MASK
                      | GDK_POINTER_MOTION_HINT_MASK);
gtk_widget_show(drawing_area);
return drawing_area;
} // end create drawing area to draw in.

GtkWidget *createVBox(GtkWidget *window)
{
GtkWidget *vbox;
vbox = gtk_vbox_new(FALSE, 2);
gtk_container_add(GTK_CONTAINER(window),vbox);
gtk_widget_show(vbox);
return vbox;
} // end create vbox for packing

gint expose_event (GtkWidget *widget, GdkEventExpose *event)
{
gdk_draw_pixmap(widget->window,
                widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                pixmap,
                event->area.x, event->area.y,
                event->area.x, event->area.y,
                event->area.width, event->area.height);

return FALSE;
}

gint configure_event (GtkWidget *widget, GdkEventConfigure *event)
{
pixmap = gdk_pixmap_new(widget->window,widget->allocation.width,widget->allocation.height,-1);
gdk_draw_rectangle(pixmap,widget->style->white_gc,TRUE,0,0,widget->allocation.width,widget->allocation.height);
return TRUE;
}

gint button_release_event(GtkWidget *widget, GdkEventButton *event)
{
GSList *nukes;
nuke *n;
game *theGame;

nukes = (GSList *)gtk_object_get_data(GTK_OBJECT(gtk_object_get_data(GTK_OBJECT(widget),"window")),"nukes");
theGame = (game *)gtk_object_get_data(GTK_OBJECT(gtk_object_get_data(GTK_OBJECT(widget),"window")),"theGame");

n = newNuke();
n->theGame = theGame;
n->darea = widget;
n->x = event->x;
n->y = event->y;
n->radius = STARTRADIUS;
n->rd = 1; // radius is expanding
n->list = nukes;
n->tickerID = gtk_timeout_add(NUKEEXPANDSPEED,(GtkFunction)expandNuke,(gpointer)n);
g_slist_append(nukes,(gpointer)n);

return TRUE;
}
